from operator import itemgetter
import sys
from typing import Any, Iterator, NamedTuple

from type_enum.utils import is_dunder, type_to_str

__all__ = ["TypeEnum"]

try:
    from collections import _tuplegetter  # undocumented implementation in C
except ImportError:

    def _tuplegetter(index: int, doc: str) -> property:
        return property(itemgetter(index), doc=doc)


class TypeEnumMeta(type):
    """Metaclass for TypeEnum."""

    def __new__(metacls, cls: str, bases: tuple[type, ...], classdict: dict[str, Any]):
        member_map: dict[str, type] = {}
        for name in classdict:
            if is_dunder(name):
                continue
            types = classdict[name]
            if isinstance(types, tuple):
                subtype = _create_tuple_class(cls, name, types)
            elif isinstance(types, dict):
                subtype = NamedTuple(name, [(k, v) for k, v in types.items()])
            else:
                raise TypeError(
                    "A TypeEnum may only contain variables of tuples or dictionaries."
                )
            classdict[name] = subtype
            member_map[name] = subtype
        classdict["_member_map"] = member_map
        try:
            exc = None
            enum_class = super().__new__(metacls, cls, bases, classdict)
        except RuntimeError as e:
            # any exceptions raised by member.__new__ will get converted to a
            # RuntimeError, so get that original exception back and raise it instead
            exc = e.__cause__ or e
        if exc is not None:
            raise exc
        return enum_class

    def __subclasscheck__(cls, subclass: type) -> bool:
        return subclass in cls._member_map.values()

    def __instancecheck__(cls, instance: object) -> bool:
        return any(isinstance(instance, typ) for typ in cls._member_map.values())

    def __iter__(cls) -> Iterator[type]:
        return cls._member_map.values()


def _create_tuple_class(basename: str, typename: str, types: tuple[type, ...]) -> type:
    """Create a new subclass of ``tuple``."""
    num_values = len(types)
    _tuple = tuple

    # ==================== methods for the new class ====================
    def __new__(cls, *args: type) -> tuple:
        if (num_args := len(args)) != num_values:
            raise TypeError(
                f"{typename}.__new__() takes {num_values} positional arguments but {num_args} were given"
            )
        return _tuple.__new__(cls, args)

    __new__.__doc__ = f"Create a new instance of {basename}.{typename}."

    repr_fmt = "(" + ", ".join("%r" for _ in range(num_values)) + ")"

    def __repr__(self) -> str:
        "Return a nicely formatted representation string"
        return basename + "." + self.__class__.__name__ + repr_fmt % self

    def __getnewargs__(self) -> tuple:
        "Return self as a plain tuple.  Used by copy and pickle."
        return _tuple(self)

    # ==================== fix `__qualname__` ====================
    for method in (__new__, __repr__, __getnewargs__):
        method.__qualname__ = f"{typename}.{method.__name__}"

    # ==================== construct the namespace of the class ====================
    field_names = tuple(f"_{i}" for i in range(num_values))

    class_namespace: dict[str, Any] = {
        "__doc__": f"{basename}.{typename}("
        + ", ".join(type_to_str(typ) for typ in types)
        + ")",
        "__slots__": (),
        "__new__": __new__,
        "__repr__": __repr__,
        "__getnewargs__": __getnewargs__,
        "__match_args__": field_names,
    }

    # ==================== add aliases for the tuple fields ====================
    # this is needed for pattern matching
    for index, name in enumerate(field_names):
        doc = sys.intern(f"Alias for field number {index}")
        class_namespace[name] = _tuplegetter(index, doc)
    return type(typename, (tuple,), class_namespace)


class TypeEnum(metaclass=TypeEnumMeta):
    pass



from types import NoneType, UnionType
from typing import Any, Literal, Union, get_args, get_origin


def type_to_str(type_: Any) -> str:
    """Convert a type to a nice string."""
    is_optional, type_ = _resolve_optional(type_)
    if type_ is None:
        return type(type_).__name__
    if type_ is Any:
        return "Any"
    if type_ is ...:
        return "..."

    if _is_literal_type(type_):
        type_ = _resolve_literal(type_)

    if hasattr(type_, "__name__"):
        name = str(type_.__name__)
    elif type_._name is None and (get_origin(type_) is not None):
        name = type_to_str(type_.__origin__)
    else:
        name = str(type_._name)

    args = get_args(type_) if hasattr(type_, "__args__") else None
    if args is None:
        ret = name
    elif name == "Callable":
        in_args_str = ", ".join(type_to_str(inner_type) for inner_type in args[0])
        out_args_str = type_to_str(args[1])
        ret = f"{name}[[{in_args_str}], {out_args_str}]"
    else:
        args_str = ", ".join(type_to_str(inner_type) for inner_type in (list(args)))
        ret = f"{name}[{args_str}]"
    if is_optional:
        return f"Optional[{ret}]"
    else:
        return ret


def _resolve_optional(type_: Any) -> tuple[bool, Any]:
    """Check whether `type_` is equivalent to `typing.Optional[T]` for some T."""
    if _is_union_type(type_):
        args = get_args(type_)
        if NoneType in args:
            optional = True
            args = tuple(a for a in args if a is not NoneType)
        else:
            optional = False
        if len(args) == 1:
            return optional, args[0]
        elif len(args) >= 2:
            return optional, Union[args]
        else:
            assert False

    if type_ is Any:
        return True, Any

    if type_ in (None, NoneType):
        return True, NoneType

    return False, type_


def _is_union_type(tp: type) -> bool:
    return tp is Union or get_origin(tp) is Union or isinstance(tp, UnionType)


def _is_literal_type(tp: type) -> bool:
    return tp is Literal or get_origin(tp) is Literal


def _resolve_literal(type_: type) -> type:
    values = get_args(type_)
    assert values
    value_types = set(type(value) for value in values)
    if len(value_types) == 1:
        return value_types.pop()
    # Constructing a Union type dynamically using a tuple is perfectly valid.
    return Union[tuple(value_types)]  # type: ignore


def is_dunder(name: str) -> bool:
    """
    Returns True if a __dunder__ name, False otherwise.
    """
    return (
        len(name) > 4
        and name[:2] == name[-2:] == "__"
        and name[2] != "_"
        and name[-3] != "_"
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['type_enum']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'type-enum',
    'version': '0.1.0',
    'description': 'Concise sum types in Python.',
    'long_description': '# `type_enum`: Concise sum types in Python\n\n### Installation\n\n```\npip install type-enum\n```\n\n### Usage\n\n```python\nfrom type_enum import TypeEnum\n\nclass BgColor(TypeEnum):\n    transparent = ()\n    name = (str,)\n    rgb = (int, int, int)\n    hsv = (int, int, int)\n\nbackground_color: BgColor = BgColor.rgb(39, 127, 168)\nassert isinstance(background_color, BgColor)\nassert not isinstance(BgColor.rgb, BgColor)  # different from Enum\n\nmatch background_color:\n    case BgColor.transparent():\n        print("no color")\n    case BgColor.name(color_name):\n        print(f"color name: {color_name}")\n    case BgColor.rgb(red, green, blue):\n        print(f"RGB: {red}, {green}, {blue}")\n    case BgColor.hsv(hue, saturation, value):\n        print(f"HSV: {hue}, {saturation}, {value}")\n# will print "RGB: 39, 127, 168"\n```\n\nYou can also specify field names by using a dictionary instead of a tuple:\n\n```python\nfrom type_enum import TypeEnum\n\nclass BgColor(TypeEnum):\n    transparent = ()\n    name = (str,)\n    rgb = {"red": int, "green": int, "blue": int}  # named args\n    hsv = {"hue": int, "saturation": int, "value": int}\n\nbackground_color = BgColor.rgb(red=39, green=127, blue=168)\nassert isinstance(background_color, BgColor)\n\nmatch background_color:\n    case BgColor.transparent():\n        print("no color")\n    case BgColor.name(color_name):\n        print(f"color name: {color_name}")\n    case BgColor.rgb(red=r, green=g, blue=b):\n        print(f"RGB: {r}, {g}, {b}")\n    case BgColor.hsv(hue=h, saturation=s, value=v):\n        print(f"HSV: {h}, {s}, {v}")\n```\n',
    'author': 'Thomas MK',
    'author_email': 'tmke8@posteo.net',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/tmke8/type_enum',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
